open Header
open Ast
open Printf (* Header for write file *)

module StringSet = Set.Make(String)

module StringMap = Map.Make(String)

(*
  User-defined Map, key: (string * int)
  For Formular Overloading
 *)

module FunctionParamLen =
   struct
     type t = string * int
     let compare (x1,x2) (y1,y2) =
       if x1 < y1 then -1
       else if x1 > y1 then 1
       else if x2 < y2 then -1
       else if x2 > y2 then 1
       else 0
   end ;;

module FPMap = Map.Make(FunctionParamLen)

(* Symbol table: Information about all the names in scope *)

type f_table = formular FPMap.t

type v_table = string list StringMap.t

type rettype = Float| Matrix

(*
   Translate program into C++ code
   @parameter : program : formulars * statements (defined in ast.ml)
   @parameter : (title, author, cpp file name, latex file name)
 *)

let translate ((formulars, statements) : program) (title, author, cppfile, lfile): unit =

  (* global float variable table *)
  let v_set = StringSet.empty in

  (* global matrix variable table *)
  let m_set = StringSet.empty in

  (* # of logical expressions *)
  let l_count = 0 in

  (* # of function calls *)
  let c_count = 0 in

  (* List of instructions in C++ *)
  let i_list = [] in

  (* List of function definitions *)
  let f_list = [] in

  (* List of function declarations *)
  let f_decl_list = [] in

  (* Global variables used in function *)
  let func_vars : v_table = StringMap.empty in

  (* Function map *)
  let func_decls : f_table = List.fold_left
      (fun functions func -> ignore(List.fold_left (fun param_map expr -> match expr with
                                                                           Id(n) -> if StringMap.mem n param_map then raise (Failure ("Parameter redefinition of " ^ n)) else
                                                                            (StringMap.add n 0 param_map)
                                                                           | _ -> raise (Failure ("Parameter is not an ID"))
                                                   )
      StringMap.empty func.parameter);
      if FPMap.mem (func.fname, List.length func.parameter) functions then raise (Failure ("Function " ^ func.fname ^ " with " ^ string_of_int (List.length func.parameter) ^ " parameters already exists."))
      else FPMap.add (func.fname, List.length func.parameter) func functions)
      FPMap.empty formulars
  in

  let func_name = FPMap.fold (fun k v set -> StringSet.add (fst(k)) set) func_decls StringSet.empty
  in

  (* Evaluation inside a function *)
  let rec formular_eval (v_set, g_set) (expr : expr) =
  (match expr with
    Call(f, para_list) -> let fdecl =
                            try
                                FPMap.find (f, List.length para_list) func_decls
                            with Not_found -> raise (Failure ("Undefined function " ^ f))
                            in
                            if (List.length fdecl.parameter)<>(List.length para_list) then raise (Failure ("Unmatched parameters"))
                            else
                            let (v_set, g_set) = List.fold_left
                                (fun (v_set, g_set) actual ->
                                  let v, (v_set, g_set) = formular_eval (v_set, g_set) actual in
                                  match v with
                                    Float -> (v_set, g_set)
                                  | Matrix -> raise (Failure ("Formular can not take in matrix"))
                                )
                                (v_set, g_set) (List.rev para_list)
                            in
                            Float, (v_set, g_set)
  | Binop(e1, op, e2) ->  let v1, (v_set, g_set) = formular_eval (v_set, g_set) e1 in
                          let v2, (v_set, g_set) = formular_eval (v_set, g_set) e2 in
                          (match (v1,v2) with
                              (Float, Float) ->
                                              (match op with
                                                Equal -> Float, (v_set, g_set)
                                              | Neq -> Float, (v_set, g_set)
                                              | Less -> Float, (v_set, g_set)
                                              | Leq -> Float, (v_set, g_set)
                                              | Greater -> Float, (v_set, g_set)
                                              | Geq -> Float, (v_set, g_set)
                                              | Add -> Float, (v_set, g_set)
                                              | Sub -> Float, (v_set, g_set)
                                              | Mul -> Float, (v_set, g_set)
                                              | Div -> Float, (v_set, g_set)
                                              | Pow -> Float, (v_set, g_set)
                                              | _ -> raise (Failure ("Matirx op can't apply on Float"))
                                              )
                          |   _ -> raise (Failure ("No matrix in formular"))
                          )
  | Sum(a, u, e) -> let vid, ex = match a with
                        Assign(vid, ex) -> vid, ex
                      | _ -> raise (Failure ("first parameter in sum should be assign"))
                     in
                     let v1, (v_set, g_set) = formular_eval (v_set, g_set) ex in
                     let v2, (v_set, g_set) = formular_eval (v_set, g_set) u in
                     let tmp_v_set = if StringSet.mem vid v_set then v_set else StringSet.add vid v_set in
                     let v3, (tmp_v_set, g_set) = formular_eval (tmp_v_set, g_set) e in
                     (match (v1,v2,v3) with
                        (Float, Float, Float) -> let v_set = if StringSet.mem vid v_set then tmp_v_set else StringSet.remove vid tmp_v_set in
                                            Float, (v_set, g_set)
                      | _ -> raise (Failure ("no matrix in sum"))
                     )
  | Prod(a, u, e) -> let vid, ex = match a with
                        Assign(vid, ex) -> vid, ex
                      | _ -> raise (Failure ("first parameter in prod should be assign"))
                     in
                     let v1, (v_set, g_set) = formular_eval (v_set, g_set) ex in
                     let v2, (v_set, g_set) = formular_eval (v_set, g_set) u in
                     let tmp_v_set = if StringSet.mem vid v_set then v_set else StringSet.add vid v_set in
                     let v3, (tmp_v_set, g_set) = formular_eval (tmp_v_set, g_set) e in
                     (match (v1,v2,v3) with
                        (Float, Float, Float) -> let v_set = if StringSet.mem vid v_set then tmp_v_set else StringSet.remove vid tmp_v_set in
                                            Float, (v_set, g_set)
                      | _ -> raise (Failure ("no matrix in prod"))
                     )
  | Uniop(op, e) -> let v, (v_set, g_set) = formular_eval (v_set, g_set) e in
                    ( match v with
                        Float -> (
                                  match op with
                                    Sin -> Float, (v_set, g_set)
                                  | Cos -> Float, (v_set, g_set)
                                  | Log -> Float, (v_set, g_set)
                                  | Tan -> Float, (v_set, g_set)
                                  | _   -> raise (Failure ("Un-supported float unit operator"))
                                 )
                      | _   ->  raise (Failure ("No matrix in formular"))
                    )
  | Literal(f) -> Float, (v_set, g_set)
  | Id(s) -> if StringSet.mem s v_set then
              Float, (v_set, g_set)
              else let g_set = StringSet.add s g_set in
              Float, (v_set, g_set)
  | _ -> raise (Failure ("Invalid expression in formular"))
  )
  in

  let process_formular (func_vars, f_list, f_decl_list) (fdecl : formular) =

    let get_name (name:expr) = match name with
        Id(n) -> n
      |   _ -> raise (Failure ("Parameter is not an ID"))
    in

    let local_set = List.fold_left
      (fun set expr -> StringSet.add (get_name expr) set)
      StringSet.empty fdecl.parameter
    in

    let process_case (v_set, g_set) (case : case) = let v, (v_set, g_set) = formular_eval (v_set, g_set) case.expression in
                                                    let v, (v_set, g_set) = formular_eval (v_set, g_set) case.condition in
                                                    let ret = "if(" ^ fst(expr_to_string_c case.condition) ^ ") return "
                                                              ^ fst(expr_to_string_c case.expression) ^ ";" in
                                                    (v_set, g_set), ret
    in


    match fdecl.definition with
      Regular(e) -> let v, (v_set, g_set) = formular_eval (local_set, StringSet.empty) e in
                    let func_vars = StringMap.add fdecl.fname (StringSet.elements g_set) func_vars in
                    let decl = "double " ^ fdecl.fname ^ "(" ^ String.concat ", " (List.map (fun e -> "double " ^ fst(expr_to_string_c e)) fdecl.parameter) ^ ");" in
                    let ret = "double " ^ fdecl.fname ^ "(" ^ String.concat ", " (List.map (fun e -> "double " ^ fst(expr_to_string_c e)) fdecl.parameter) ^ "){\n\t"
                    ^ "return " ^ fst(expr_to_string_c e) ^ ";\n}" in
                    func_vars, ret::f_list, decl::f_decl_list

    | Piecewise(case_list) -> let (v_set, g_set), tmp = List.fold_left (fun ((v_set, g_set), ret) case ->
                                                          let (v_set, g_set), r = process_case (v_set, g_set) case in
                                                              (v_set, g_set), ret ^ "\n\t" ^ r)
                                                              ((local_set, StringSet.empty), "") case_list
                              in
                              let func_vars = StringMap.add fdecl.fname (StringSet.elements g_set) func_vars in
                              let decl = "double " ^ fdecl.fname ^ "(" ^ String.concat ", " (List.map (fun e -> "double " ^ fst(expr_to_string_c e)) fdecl.parameter) ^ ");" in
                              let ret = "double " ^ fdecl.fname ^ "(" ^  String.concat ", " (List.map (fun e -> "double " ^ fst(expr_to_string_c e)) fdecl.parameter) ^ "){"
                                ^ tmp ^ "\n\tthrow std::runtime_error(\"Illegal parameter in piecewise function " ^ fdecl.fname ^ "\");\n}" in
                                func_vars, ret::f_list, decl::f_decl_list

  in

  let (func_vars, f_list, f_decl_list) =
          List.fold_left process_formular (func_vars, f_list, f_decl_list) formulars
  in
  (* Normal evaluation *)
  let rec eval (v_set, m_set, l_count, c_count, i_list) (expr : expr) =
  (match expr with
    Call(f, para_list) -> let fdecl =
                            try
                                FPMap.find (f, List.length para_list) func_decls
                            with Not_found -> raise (Failure ("Undefined function " ^ f))
                            in
                            let g_list = StringMap.find f func_vars
                            in
                            List.iter (fun g -> if StringSet.mem g v_set then () else raise (Failure ("Uninitialized global variable " ^ g)) ) g_list;
                            if (List.length fdecl.parameter)<>(List.length para_list) then raise (Failure ("Unmatched parameters"))
                            else
                            let (v_set, m_set, l_count, c_count, i_list) = List.fold_left
                                (fun (v_set, m_set, l_count, c_count, i_list) actual ->
                                  let v, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) actual in
                                  match v with
                                    Float -> (v_set, m_set, l_count, c_count, i_list)
                                  | Matrix -> raise (Failure ("Formular can not take in matrix"))
                                )
                                (v_set, m_set, l_count, c_count, i_list) (List.rev para_list)
                            in
                            Float, (v_set, m_set, l_count, c_count, i_list)
  | Binop(e1, op, e2) ->  let v1, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) e1 in
                          let v2, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) e2 in
                          (match (v1,v2) with
                              (Float, Float) ->
                                              (match op with
                                                Equal -> Float, (v_set, m_set, l_count, c_count, i_list)
                                              | Neq -> Float, (v_set, m_set, l_count, c_count, i_list)
                                              | Less -> Float, (v_set, m_set, l_count, c_count, i_list)
                                              | Leq -> Float, (v_set, m_set, l_count, c_count, i_list)
                                              | Greater -> Float, (v_set, m_set, l_count, c_count, i_list)
                                              | Geq -> Float, (v_set, m_set, l_count, c_count, i_list)
                                              | Add -> Float, (v_set, m_set, l_count, c_count, i_list)
                                              | Sub -> Float, (v_set, m_set, l_count, c_count, i_list)
                                              | Mul -> Float, (v_set, m_set, l_count, c_count, i_list)
                                              | Div -> Float, (v_set, m_set, l_count, c_count, i_list)
                                              | Pow -> Float, (v_set, m_set, l_count, c_count, i_list)
                                              | _ -> raise (Failure ("Matirx op can't apply on Float"))
                                              )
                          |   (Float, Matrix) ->
                                              (match op with
                                                Mul -> Matrix, (v_set, m_set, l_count, c_count, i_list)
                                              | _ -> raise (Failure ("Float can only apply Mul with Matrix"))
                                              )
                          |   (Matrix, Float) -> raise (Failure ("Matirx op Float not allowed"))
                          |   (Matrix, Matrix) ->
                                              (match op with
                                                Mul -> Matrix, (v_set, m_set, l_count, c_count, i_list)
                                              | DotMul -> Matrix, (v_set, m_set, l_count, c_count, i_list)
                                              | Add -> Matrix, (v_set, m_set, l_count, c_count, i_list)
                                              | Sub -> Matrix, (v_set, m_set, l_count, c_count, i_list)
                                              | _ -> raise (Failure ("Undefined Matrix operation"))
                                              )
                          )
  | Sum(a, u, e) -> let vid, ex = match a with
                        Assign(vid, ex) -> vid, ex
                      | _ -> raise (Failure ("first parameter in sum should be assign"))
                     in
                     let v1, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) ex in
                     let v2, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) u in
                     let tmp_v_set = if StringSet.mem vid v_set then v_set else StringSet.add vid v_set in
                     let v3, (tmp_v_set, m_set, l_count, c_count, i_list) = eval (tmp_v_set, m_set, l_count, c_count, i_list) e in
                     (match (v1,v2,v3) with
                        (Float, Float, Float) -> let v_set = if StringSet.mem vid v_set then tmp_v_set else StringSet.remove vid tmp_v_set in
                                            Float, (v_set, m_set, l_count, c_count, i_list)
                      | _ -> raise (Failure ("no matrix in sum"))
                     )
  | Prod(a, u, e) -> let vid, ex = match a with
                        Assign(vid, ex) -> vid, ex
                      | _ -> raise (Failure ("first parameter in prod should be assign"))
                     in
                     let v1, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) ex in
                     let v2, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) u in
                     let tmp_v_set = if StringSet.mem vid v_set then v_set else StringSet.add vid v_set in
                     let v3, (tmp_v_set, m_set, l_count, c_count, i_list) = eval (tmp_v_set, m_set, l_count, c_count, i_list) e in
                     (match (v1,v2,v3) with
                        (Float, Float, Float) -> let v_set = if StringSet.mem vid v_set then tmp_v_set else StringSet.remove vid tmp_v_set in
                                            Float, (v_set, m_set, l_count, c_count, i_list)
                      | _ -> raise (Failure ("no matrix in prod"))
                     )
  | Literal(f) -> Float, (v_set, m_set, l_count, c_count, i_list)
  | MLiteral(m) -> Matrix, (v_set, m_set, l_count, c_count, i_list)
  | Id(s) -> if StringSet.mem s v_set then
              Float, (v_set, m_set, l_count, c_count, i_list)
              else if StringSet.mem s m_set then
              Matrix, (v_set, m_set, l_count, c_count, i_list)
              else raise (Failure ("Undeclared identifier " ^ s))
  | Uniop(op, e) -> let v, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) e in
                    (match v with
                        Matrix ->
                              ( match op with
                                Trans -> Matrix, (v_set, m_set, l_count, c_count, i_list)
                              | _ -> raise (Failure ("Un-supported matrix unit operator"))
                              )
                      | Float ->
                              ( match op with
                                  Sin -> Float, (v_set, m_set, l_count, c_count, i_list)
                                | Cos -> Float, (v_set, m_set, l_count, c_count, i_list)
                                | Log -> Float, (v_set, m_set, l_count, c_count, i_list)
                                | Tan -> Float, (v_set, m_set, l_count, c_count, i_list)
                                | _   -> raise (Failure ("Un-supported float unit operator"))
                              )
                    )
  | Assign(vid, e) ->
    if StringSet.mem vid func_name then raise (Failure ("Cannot assign a new var with the same name of a formular"))
    else
    let v, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) e in
                        (match v with
                          Float -> if StringSet.mem vid m_set then raise (Failure ("Cannot assign a new var with the same name of a matrix")) else
                                    let v_set = StringSet.add vid v_set in
                                    Float, (v_set, m_set, l_count, c_count, i_list)

                        | Matrix -> if StringSet.mem vid v_set then raise (Failure ("Cannot assign a var with a matrix")) else
                                    let m_set = StringSet.add vid m_set in
                                    Matrix, (v_set, m_set, l_count, c_count, i_list)
                        )
  )
  in

  (* Capture function calls and logical validations in top-level and handle printing *)
  let eval_top (v_set, m_set, l_count, c_count, i_list) (expr : expr) =
  (match expr with
      Call(f, para_list) -> let fdecl =
                            try
                                FPMap.find (f, List.length para_list) func_decls
                            with Not_found -> raise (Failure ("Undefined function " ^ f))
                            in
                            let g_list = StringMap.find f func_vars
                            in
                            List.iter (fun g -> if StringSet.mem g v_set then () else raise (Failure ("Uninitialized global variable " ^ g)) ) g_list;
                            if (List.length fdecl.parameter)<>(List.length para_list) then raise (Failure ("Unmatched parameters"))
                            else
                            let (v_set, m_set, l_count, c_count, i_list) = List.fold_left
                                (fun (v_set, m_set, l_count, c_count, i_list) actual ->
                                  let v, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) actual in
                                  match v with
                                    Float -> (v_set, m_set, l_count, c_count, i_list)
                                  | Matrix -> raise (Failure ("Formular can not take in matrix"))
                                )
                                (v_set, m_set, l_count, c_count, i_list) (List.rev para_list)
                            in

                            let display = if (List.length para_list)>0 then
                            "string(\"\")" ^ "+\"" ^  f ^ "(" ^ String.escaped(expr_list_to_string para_list) ^ ")=" ^ f ^ "(\"+"
                                ^ (expr_list_to_string_p "+\", \"+" para_list) ^ "+\")=\"+dtos(" ^ f ^ "("
                                ^ (expr_list_to_string_c para_list) ^ "))"
                            else "string(\"\")" ^ "+\"" ^  f ^ "(" ^ String.escaped(expr_list_to_string para_list) ^ ")=" ^ f ^ "(\""
                                ^ (expr_list_to_string_p "+\", \"+" para_list) ^ "+\")=\"+dtos(" ^ f ^ "("
                                ^ (expr_list_to_string_c para_list) ^ "))"
                            in
                            let c_count = c_count + 1
                            in
                            let instruction = "c_result[" ^ (string_of_int (c_count-1)) ^ "]=" ^ display ^ ";"
                            in
                            let i_list = instruction :: i_list
                            in
                            (v_set, m_set, l_count, c_count, i_list)

    | Binop(e1, op, e2) -> let v1, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) e1 in
                           let v2, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) e2 in
                           (match (v1,v2) with
                              (Float, Float) -> let display_first = "string(\"\")" ^ "+\"" ^ String.escaped(fst(expr_to_string expr)) ^ " \"" in
                                              let display_left = "dtos(" ^ fst(expr_to_string_c e1) ^ ")" in
                                              let display_right = "dtos(" ^ fst(expr_to_string_c e2) ^ ")" in
                                              let display_last = "(" ^ fst(expr_to_string_c expr) ^ "?" ^ "\" true\":\" false\"" ^ ")" in
                                              (match op with
                                                Equal -> let display = display_first ^ "+" ^ "\"\\\\Rightarrow \"" ^ "+" ^ display_left ^ "+\"==\"+" ^ display_right ^ "+" ^ "\" \\\\Rightarrow\"" ^ "+" ^ display_last in
                                                          let l_count = l_count + 1 in
                                                          let instruction = "l_result[" ^ (string_of_int (l_count-1)) ^ "]=" ^ display ^ ";" in
                                                          let i_list = instruction :: i_list in
                                                          (v_set, m_set, l_count, c_count, i_list)
                                              | Neq -> let display = display_first ^ "+" ^ "\"\\\\Rightarrow \"" ^ "+" ^ display_left ^ "+\"!=\"+" ^ display_right ^ "+" ^ "\" \\\\Rightarrow\"" ^ "+" ^ display_last in
                                                          let l_count = l_count + 1 in
                                                          let instruction = "l_result[" ^ (string_of_int (l_count-1)) ^ "]=" ^ display ^ ";" in
                                                          let i_list = instruction :: i_list in
                                                          (v_set, m_set, l_count, c_count, i_list)
                                              | Less -> let display = display_first ^ "+" ^ "\"\\\\Rightarrow \"" ^ "+" ^ display_left ^ "+\"<\"+" ^ display_right ^ "+" ^ "\" \\\\Rightarrow\"" ^ "+" ^ display_last in
                                                          let l_count = l_count + 1 in
                                                          let instruction = "l_result[" ^ (string_of_int (l_count-1)) ^ "]=" ^ display ^ ";" in
                                                          let i_list = instruction :: i_list in
                                                          (v_set, m_set, l_count, c_count, i_list)
                                              | Leq -> let display = display_first ^ "+" ^ "\"\\\\Rightarrow \"" ^ "+" ^ display_left ^ "+\"<=\"+" ^ display_right ^ "+" ^ "\" \\\\Rightarrow\"" ^ "+" ^ display_last in
                                                          let l_count = l_count + 1 in
                                                          let instruction = "l_result[" ^ (string_of_int (l_count-1)) ^ "]=" ^ display ^ ";" in
                                                          let i_list = instruction :: i_list in
                                                          (v_set, m_set, l_count, c_count, i_list)
                                              | Greater -> let display = display_first ^ "+" ^ "\"\\\\Rightarrow \"" ^ "+" ^ display_left ^ "+\">\"+" ^ display_right ^ "+" ^ "\" \\\\Rightarrow\"" ^ "+" ^ display_last in
                                                          let l_count = l_count + 1 in
                                                          let instruction = "l_result[" ^ (string_of_int (l_count-1)) ^ "]=" ^ display ^ ";" in
                                                          let i_list = instruction :: i_list in
                                                          (v_set, m_set, l_count, c_count, i_list)
                                              | Geq -> let display = display_first ^ "+" ^ "\"\\\\Rightarrow \"" ^ "+" ^ display_left ^ "+\">=\"+" ^ display_right ^ "+" ^ "\" \\\\Rightarrow\"" ^ "+" ^ display_last in
                                                          let l_count = l_count + 1 in
                                                          let instruction = "l_result[" ^ (string_of_int (l_count-1)) ^ "]=" ^ display ^ ";" in
                                                          let i_list = instruction :: i_list in
                                                          (v_set, m_set, l_count, c_count, i_list)
                                              | Add -> let instruction = fst(expr_to_string_c expr) ^ ";" in
                                                        let i_list = instruction :: i_list in
                                                          (v_set, m_set, l_count, c_count, i_list)
                                              | Sub -> let instruction = fst(expr_to_string_c expr) ^ ";" in
                                                        let i_list = instruction :: i_list in
                                                          (v_set, m_set, l_count, c_count, i_list)
                                              | Mul -> let instruction = fst(expr_to_string_c expr) ^ ";" in
                                                        let i_list = instruction :: i_list in
                                                          (v_set, m_set, l_count, c_count, i_list)
                                              | Div -> let instruction = fst(expr_to_string_c expr) ^ ";" in
                                                        let i_list = instruction :: i_list in
                                                          (v_set, m_set, l_count, c_count, i_list)
                                              | Pow -> let instruction = fst(expr_to_string_c expr) ^ ";" in
                                                        let i_list = instruction :: i_list in
                                                          (v_set, m_set, l_count, c_count, i_list)
                                              | _ -> raise (Failure ("Matirx op can't apply on Float"))
                                              )
                          |   (Float, Matrix) ->
                                              (match op with
                                                Mul -> let instruction = fst(expr_to_string_c expr) ^ ";" in
                                                        let i_list = instruction :: i_list in
                                                          (v_set, m_set, l_count, c_count, i_list)
                                              | _ -> raise (Failure ("Float can only apply Mul with Matrix"))
                                              )
                          |   (Matrix, Float) -> raise (Failure ("Matirx op Float not allowed"))
                          |   (Matrix, Matrix) ->
                                              (match op with
                                                Mul -> let instruction = fst(expr_to_string_c expr) ^ ";" in
                                                        let i_list = instruction :: i_list in
                                                          (v_set, m_set, l_count, c_count, i_list)
                                              | DotMul -> let instruction = fst(expr_to_string_c expr) ^ ";" in
                                                        let i_list = instruction :: i_list in
                                                          (v_set, m_set, l_count, c_count, i_list)
                                              | Add -> let instruction = fst(expr_to_string_c expr) ^ ";" in
                                                        let i_list = instruction :: i_list in
                                                         (v_set, m_set, l_count, c_count, i_list)
                                              | Sub -> let instruction = fst(expr_to_string_c expr) ^ ";" in
                                                        let i_list = instruction :: i_list in
                                                         (v_set, m_set, l_count, c_count, i_list)
                                              | _ -> raise (Failure ("Undefined Matrix operation"))
                                              )
                          )
    | Sum(a, u, e) -> let v, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) expr in
                        let instruction = fst(expr_to_string_c expr) ^ ";" in
                        let i_list = instruction :: i_list in
                        (v_set, m_set, l_count, c_count, i_list)
    | Prod(a, u, e) -> let v, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) expr in
                        let instruction = fst(expr_to_string_c expr) ^ ";" in
                        let i_list = instruction :: i_list in
                        (v_set, m_set, l_count, c_count, i_list)
    | Literal(f) -> (v_set, m_set, l_count, c_count, i_list)
    | MLiteral(m) -> (v_set, m_set, l_count, c_count, i_list)
    | Id(s) -> (v_set, m_set, l_count, c_count, i_list)
    | Uniop(op, e) -> let v, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) e in
                        (match v with
                          Matrix ->
                                  ( match op with
                                     Trans ->
                                         let instruction = fst(expr_to_string_c expr) ^ ";" in
                                          let i_list = instruction :: i_list in
                                          (v_set, m_set, l_count, c_count, i_list)
                                    | _ -> raise (Failure ("Un-supported matrix unit operator"))
                                  )
                        | Float  ->
                                  (
                                    match op with
                                      Sin ->  let instruction = fst(expr_to_string_c expr) ^ ";" in
                                               let i_list = instruction :: i_list in
                                                (v_set, m_set, l_count, c_count, i_list)
                                    | Cos ->  let instruction = fst(expr_to_string_c expr) ^ ";" in
                                               let i_list = instruction :: i_list in
                                                (v_set, m_set, l_count, c_count, i_list)
                                    | Log ->  let instruction = fst(expr_to_string_c expr) ^ ";" in
                                               let i_list = instruction :: i_list in
                                                (v_set, m_set, l_count, c_count, i_list)
                                    | Tan ->  let instruction = fst(expr_to_string_c expr) ^ ";" in
                                               let i_list = instruction :: i_list in
                                                (v_set, m_set, l_count, c_count, i_list)
                                    | _   ->  raise (Failure ("Un-supported float unit operator"))
                                  )
                        )
    | Assign(vid, e) -> let v, (v_set, m_set, l_count, c_count, i_list) = eval (v_set, m_set, l_count, c_count, i_list) e in
                        (match v with
                          Float -> if StringSet.mem vid m_set then raise (Failure ("Cannot assign a new var with the same name of a matrix")) else
                                    let v_set = StringSet.add vid v_set in
                                    let instruction = fst(expr_to_string_c expr) ^ ";" in
                                    let i_list = instruction :: i_list in
                                    (v_set, m_set, l_count, c_count, i_list)
                        | Matrix -> if StringSet.mem vid v_set then raise (Failure ("Cannot assign a var with a matrix")) else
                                    let m_set = StringSet.add vid m_set in
                                    let instruction = fst(expr_to_string_c expr) ^ ";" in
                                    let i_list = instruction :: i_list in
                                    (v_set, m_set, l_count, c_count, i_list)
                        )
  )
  in

  let process_stmt (v_set, m_set, l_count, c_count, i_list) (statement : statement) =
    match statement with
    Seq(expr_list) -> List.fold_left
      (fun a b -> eval_top a b) (v_set, m_set, l_count, c_count, i_list) expr_list
  in

  (* Genereate variable set,
               matrix set,
               logic expression count,
               formular evaluation count,
               instruction list
  *)
  let (v_set, m_set, l_count, c_count, i_list) =
          List.fold_left process_stmt (v_set, m_set, l_count, c_count, i_list) statements
  in

  let v_list = StringSet.elements(v_set) and m_list = StringSet.elements(m_set)

  in

  let fdef_latex_l =
     List.rev (FPMap.fold (fun k v l -> String.escaped(formular_to_string v):: l) func_decls [])
  in

  let (v_str, c1) = List.fold_left
                (fun (str, i) var -> (str ^ "double &" ^ var ^ " = vdata[" ^ string_of_int i ^ "];\n", i+1)) ("", 0) v_list
  and (m_str, c2) = List.fold_left
                (fun (str, i) var -> (str ^ "matrix &" ^ var ^ " = mdata[" ^ string_of_int i ^ "];\n", i+1)) ("", 0) m_list
  in

  (* Combine the strings *)
  let code =
    header v_list m_list l_count c_count fdef_latex_l title author date ^ "\n" ^

    String.concat "\n\n" (List.rev f_decl_list) ^ "\n" ^
    matrix_class ^ "\n" ^

    "double vdata[NUM_OF_VARIABLES];\n" ^

    "matrix mdata[NUM_OF_MATRIX_VARIABLES];\n" ^

    v_str ^ "\n" ^
    m_str ^ "\n" ^

    dtos ^ "\n" ^
    String.concat "\n\n" (List.rev f_list) ^ "\n" ^
    main_preamble v_list m_list fdef_latex_l ^ "\n" ^

    String.concat "\n" (List.map (fun ins -> "\t" ^ ins) (List.rev i_list)) ^ "\n" ^
    latex_print lfile
  in

  (* Write string to file *)
  let out  = open_out cppfile in
    fprintf out "%s\n" code;
    close_out out;


