open Ast
open Header

module StringMap = Map.Make(String)
module StringSet = Set.Make(String)

(*
  User-defined Map, key: (string * int)
  For Formular Overloading
 *)

module FunctionParamLen =
   struct
     type t = string * int
     let compare (x1,x2) (y1,y2) =
       if x1 < y1 then -1
       else if x1 > y1 then 1
       else if x2 < y2 then -1
       else if x2 > y2 then 1
       else 0
   end ;;

module FPMap = Map.Make(FunctionParamLen)

(* Symbol table: Information about all the names in scope *)
type v_table = float StringMap.t

type f_table = formular FPMap.t

type m_table = matrix StringMap.t

type global_config = v_table * m_table

type rettype = Float of float | Matrix of matrix

(* Global mutable list,
    information about formualr evalution instrcution
                and   logical evalution instruction
  *)
let fcall_strs = ref [];;
let lvali_strs = ref [];;

(* Get item in tripe-pairs *)
let first (x,_,_) = x;;
let second (_,y,_) = y;;
let third (_,_,z) = z;;

(* Matrix Multiplication *)
let matrix_constant_mul (lambda:float) (x:matrix)  : matrix =
    let x_row = Array.length x in
    if x_row = 0 then raise (Failure ("Invalid matrix with row# = 0")) else let x_col = Array.length x.(0) in
    if x_col = 0 then raise (Failure ("Invalid matrix with col# = 0")) else
    let z = Array.make_matrix x_row x_col 0. in
    for i = 0 to x_row-1 do
        for j = 0 to x_col-1 do
            z.(i).(j) <- lambda *. x.(i).(j)
        done
    done;
    (z)

(* Matrix Multiplication *)
let matrix_multiply (x:matrix) (y:matrix) : matrix =
  let x_row = Array.length x and y_row = Array.length y in
  if x_row = 0 then raise (Failure ("Invalid matrix with row# = 0")) else let x_col = Array.length x.(0) in
  if y_row = 0 then raise (Failure ("Invalid matrix with row# = 0")) else let y_col = Array.length y.(0) in
  if (x_col = 0) || (y_col = 0) then raise (Failure ("Invalid matrix with col# = 0")) else
  if x_col <> y_row then raise (Failure ("Invalid matrix multiply ")) else
  let z = Array.make_matrix x_row y_col 0. in
  for i = 0 to x_row-1 do
    for j = 0 to y_col-1 do
      for k = 0 to y_row-1 do
        z.(i).(j) <- z.(i).(j) +. x.(i).(k) *. y.(k).(j)
      done
    done
  done;
  (z)

(* Matrix Multiplication *)
let matrix_dot_multiply (x:matrix) (y:matrix) : matrix =
  let x_row = Array.length x and y_row = Array.length y in
  if x_row = 0 then raise (Failure ("Invalid matrix with row# = 0")) else let x_col = Array.length x.(0) in
  if y_row = 0 then raise (Failure ("Invalid matrix with row# = 0")) else let y_col = Array.length y.(0) in
  if (x_col = 0) || (y_col = 0) then raise (Failure ("Invalid matrix with col# = 0")) else
  if (x_row <> y_row) || (x_col <> y_col) then raise (Failure ("Invalid matrix dot multiply ")) else
  let z = Array.make_matrix x_row x_col 0. in
  for i = 0 to x_row-1 do
    for j = 0 to x_col-1 do
        z.(i).(j) <- x.(i).(j) *. y.(i).(j)
    done
  done;
  (z)

(* Matrix addition *)
let matrix_add (x:matrix) (y:matrix) : matrix =
  let x_row = Array.length x and y_row = Array.length y in
  if x_row = 0 then raise (Failure ("Invalid matrix with row# = 0")) else let x_col = Array.length x.(0) in
  if y_row = 0 then raise (Failure ("Invalid matrix with row# = 0")) else let y_col = Array.length y.(0) in
  if (x_col = 0) || (y_col = 0) then raise (Failure ("Invalid matrix with col# = 0")) else
  if (x_row <> y_row) || (x_col <> y_col) then raise (Failure ("Invalid matrix dot multiply ")) else
  let z = Array.make_matrix x_row x_col 0. in
  for i = 0 to x_row-1 do
    for j = 0 to x_col-1 do
        z.(i).(j) <- x.(i).(j) +. y.(i).(j)
    done
  done;
  (z)

(* Matrix subtraction *)
let matrix_sub (x:matrix) (y:matrix) : matrix =
  let x_row = Array.length x and y_row = Array.length y in
  if x_row = 0 then raise (Failure ("Invalid matrix with row# = 0")) else let x_col = Array.length x.(0) in
  if y_row = 0 then raise (Failure ("Invalid matrix with row# = 0")) else let y_col = Array.length y.(0) in
  if (x_col = 0) || (y_col = 0) then raise (Failure ("Invalid matrix with col# = 0")) else
  if (x_row <> y_row) || (x_col <> y_col) then raise (Failure ("Invalid matrix dot multiply ")) else
  let z = Array.make_matrix x_row x_col 0. in
  for i = 0 to x_row-1 do
    for j = 0 to x_col-1 do
        z.(i).(j) <- x.(i).(j) -. y.(i).(j)
    done
  done;
  (z)

(* Matrix Transpose *)
let matrix_transpose (x:matrix) : matrix =
  let x_row = Array.length x in
  if x_row = 0 then raise (Failure ("Invalid matrix with row# = 0")) else let x_col = Array.length x.(0) in
  if x_col = 0 then raise (Failure ("Invalid matrix with col# = 0")) else
  let z = Array.make_matrix x_col x_row 0. in
  for i = 0 to x_col-1 do
    for j = 0 to x_row-1 do
        z.(i).(j) <- x.(j).(i)
    done
  done;
  (z)

(*
   Executing program, output result in stdout
   @parameter : program : formulars * statements (defined in ast.ml)
   @parameter : (title, author)
*)

let run ((formulars, statements) : program) (title, author) : unit =

  (* Build a symbol table for function declarations *)
  let func_decls : f_table = List.fold_left
      (* Semantics checking: parameter names should be unique; function with same signature (name + # of parameters) should be unique *)
      (fun functions func -> ignore(List.fold_left (fun param_map expr -> match expr with
                                                                           Id(n) -> if StringSet.mem n param_map then raise (Failure ("Parameter redefinition of " ^ n)) else
                                                                            (StringSet.add n param_map)
                                                                           | _ -> raise (Failure ("Parameter is not an ID"))
                                                   )
      StringSet.empty func.parameter);
      if FPMap.mem (func.fname, List.length func.parameter) functions then raise (Failure ("Function " ^ func.fname ^ " with " ^ string_of_int (List.length func.parameter) ^ " parameters already exists."))
      else FPMap.add (func.fname, List.length func.parameter) func functions)
      FPMap.empty formulars
  in

  let func_name = FPMap.fold (fun k v set -> StringSet.add (fst(k)) set) func_decls StringSet.empty
  in

  (* Initial Empty Matrix table, id(string) -> val(matrix) *)
  let global_matrices : m_table = StringMap.empty
  in

  (* Initial Empty Variable table, id(string) -> val(Float) *)
  let global_vars : v_table = StringMap.empty
  in

  (* Formular execution utilities *)
  let rec exec_formular (fdecl : formular)(actuals : float list)(global_vars : v_table) : rettype * v_table =
  	let get_name (name:expr) = match name with
  			Id(n) -> n
  		|   _ -> raise (Failure ("Parameter is not an ID"))
  	in

	let find_case (config : rettype * v_table * bool) (case : case) : bool =
	let global_vars = second(config) in
  let v, (_, _) = eval (global_vars, global_matrices) case.condition in
  match v with
  Float(f) -> (f > 0.)
  | _ -> raise (Failure ("Matrix is not allowed in case condition"))

    in

  	let local_vars : v_table = List.fold_left2 (fun local_vars actual name -> StringMap.add (get_name name) actual local_vars) global_vars actuals fdecl.parameter
  	in
  	match fdecl.definition with
  			Regular(e) -> let (result, (local_vars, global_matrices)) = eval (local_vars, global_matrices) e in (result, global_vars) (* Note: global_matrix is empty *)
		|   Piecewise(case_list) -> let (result, local_vars, is_found) = List.fold_left
				(fun (config : rettype * v_table * bool) (case : case) -> if (not (third(config)) && find_case config case) then (fst(eval (local_vars, global_matrices) case.expression), second(config), true) else (config) )
				(Float(0.), local_vars, false) case_list
			in
			if is_found then (result, global_vars) else raise (Failure ("Piecewise not complete"))

  and

  (* Main evaluation function *)
  eval ((global_vars : v_table), (global_matrices : m_table)) (exp : expr) : rettype * global_config = match exp with
		(* Return Constant  *)
        Literal(f) -> Float(f), (global_vars, global_matrices)
  	|   MLiteral(m) -> Matrix(m), (global_vars, global_matrices)
    (* first try to find in float vars, then try to find in matrix vars *)
  	| 	Id(s) -> if StringMap.mem s global_vars then
	    Float((StringMap.find s global_vars)), (global_vars, global_matrices)
	  	else if StringMap.mem s global_matrices then
	    Matrix((StringMap.find s global_matrices)), (global_vars, global_matrices)
	  	else raise (Failure ("Undeclared identifier " ^ s))
	| 	Assign(vid, e) ->
    if StringSet.mem vid func_name then raise (Failure ("Cannot assign a new var with the same name of a formular"))
    else
		let (v, (global_vars, global_matrices)) = eval (global_vars, global_matrices) e in
  		(match v with
      		  Float(f) -> if StringMap.mem vid global_matrices then raise (Failure ("Cannot assign a new var with the same name of a matrix")) else
      					(Float(f), ((StringMap.add vid f global_vars), global_matrices))
      		| Matrix(m) -> if StringMap.mem vid global_vars then raise (Failure ("Cannot assign a var with a matrix")) else
      					           Matrix(m), (global_vars, (StringMap.add vid m global_matrices))
      )
  |   Sum(a, u, e) -> let vid, ex = match a with
                        Assign(vid, ex) -> vid, ex
                      | _ -> raise (Failure ("first parameter in sum should be assign"))
                     in
                     let v1, (global_vars, global_matrices) = eval (global_vars, global_matrices) ex in
                     let v2, (global_vars, global_matrices) = eval (global_vars, global_matrices) u in
                     (match (v1,v2) with
                        (Float(f1), Float(f2)) ->   let sum = ref 0. in
                                                    for i = (int_of_float f1) to (int_of_float f2) do
                                                    let global_vars = StringMap.add vid (float_of_int i) global_vars in
                                                    let v3, (global_vars, global_matrices) = eval (global_vars, global_matrices) e in
                                                    let ret = match v3 with
                                                                Float(f3) -> f3
                                                              | _ -> raise (Failure ("no matrix in sum"))
                                                    in
                                                      sum := !sum +. ret
                                                    done;
                                                    Float(!sum), (global_vars, global_matrices)
                      | _ -> raise (Failure ("no matrix in sum"))
                     )
  |   Prod(a, u, e) -> let vid, ex = match a with
                        Assign(vid, ex) -> vid, ex
                      | _ -> raise (Failure ("first parameter in prod should be assign"))
                     in
                     let v1, (global_vars, global_matrices) = eval (global_vars, global_matrices) ex in
                     let v2, (global_vars, global_matrices) = eval (global_vars, global_matrices) u in
                     (match (v1,v2) with
                        (Float(f1), Float(f2)) ->   let prod = ref 1. in
                                                    for i = (int_of_float f1) to (int_of_float f2) do
                                                    let global_vars = StringMap.add vid (float_of_int i) global_vars in
                                                    let v3, (global_vars, global_matrices) = eval (global_vars, global_matrices) e in
                                                    let ret = match v3 with
                                                                Float(f3) -> f3
                                                              | _ -> raise (Failure ("no matrix in prod"))
                                                    in
                                                      prod := !prod *. ret
                                                    done;
                                                    Float(!prod), (global_vars, global_matrices)
                      | _ -> raise (Failure ("no matrix in prod"))
                     )
	|   Uniop(op, e) -> let (r, (global_vars, global_matrices)) = eval (global_vars, global_matrices) e in
											  (
                          match r with
  												  Matrix(m) ->( match op with
                                              Trans -> (Matrix(matrix_transpose m), (global_vars, global_matrices))
                                            | _    -> raise (Failure ("Cannot perform other unit operator on matrix"))
                                        )
  											  | Float(f) -> (
                                          match op with
                                              Sin -> (Float(sin f), (global_vars, global_matrices))
                                            | Cos -> ((Float(cos f), (global_vars, global_matrices)))
                                            | Log -> ((Float(log10 f), (global_vars, global_matrices)))
                                            | Tan -> ((Float(tan f), (global_vars, global_matrices)))
                                            | _ -> raise (Failure ("Unsupported unit operator on float number")) (* Future work: add more unit operator function *)
                                        )
                        )
  | 	Binop(e1, op, e2) ->
	  	    let v1, (global_vars, global_matrices) = eval (global_vars, global_matrices) e1 in
          let v2, (global_vars, global_matrices) = eval (global_vars, global_matrices) e2 in
          let boolean i = if i then 1. else 0. in
          (match (v1,v2) with
            (Float(f1),  Float(f2)) -> 		  (match op with
                    											    Add -> Float(f1 +. f2)
                    											  | Sub -> Float(f1 -. f2)
                    											  | Mul -> Float(f1 *. f2)
                    											  | Div -> if f2 = 0. then raise (Failure ("Division by 0!")) else
                                                Float(f1 /. f2)
                    											  | Equal -> Float(boolean (f1 = f2))
                    											  | Neq -> Float(boolean (f1 <> f2))
                    											  | Less -> Float(boolean (f1 < f2))
                    											  | Leq -> Float(boolean (f1 <= f2))
                    											  | Greater -> Float(boolean (f1 > f2))
                    											  | Geq -> Float(boolean (f1 >= f2))
                    										    | Pow -> Float(f1 ** f2)
                    										    | _ -> raise (Failure ("Matirx op can't apply on float"))
                    										    ), (global_vars, global_matrices)
            | (Float(f1), Matrix(m1)) -> (match op with
                          									 Mul -> (Matrix(matrix_constant_mul f1 m1 ), (global_vars, global_matrices))
                          								 | _ -> raise (Failure ("Float can only apply Mul with matrix"))
                        								 )
            | (Matrix(m1), Float(f1)) ->  raise (Failure ("Matirx op Float not allowed"))
            | (Matrix(m1), Matrix(m2)) -> (match op with
                                              Mul -> (Matrix(matrix_multiply m1 m2), (global_vars, global_matrices)) (* Matrix Multiplication operation*)
                          									| DotMul -> (Matrix(matrix_dot_multiply m1 m2), (global_vars, global_matrices)) (* Matrix DotMul operation*)
                                            | Add -> (Matrix(matrix_add m1 m2), (global_vars, global_matrices)) (* Matrix add operation*)
                                            | Sub -> (Matrix(matrix_sub m1 m2), (global_vars, global_matrices)) (* Matrix sub operation*)
                          									| _ -> raise (Failure ("Matirx op Matrix not allowed"))
                        								  )
         )

	|   Call(f, para_list) ->
			let fdecl =
			try
	    		FPMap.find (f, List.length para_list) func_decls
	    with Not_found -> raise (Failure ("Undefined function " ^ f))
	    in
	      let (actuals: float list), (global_vars, global_matrices) = List.fold_left
	      		(
            fun (actuals, (global_vars, global_matrices)) actual ->
      				let v, (global_vars, global_matrices) = eval (global_vars, global_matrices) actual in
      				 match v with
      				   Float(f) -> f :: actuals, (global_vars, global_matrices)
      				 | Matrix(m) -> raise (Failure ("Formular can not take in matrix"))
				    )
   	      	([], (global_vars, global_matrices)) (List.rev para_list)
	  		in let (r, v) = exec_formular fdecl actuals global_vars
      in let f = match r with Matrix(m) -> raise (Failure ("Matirx not allowed in function return type")) | Float(f) -> f
      in (Float(f), (v, global_matrices))

  in

  (* Capture function calls and logical validations in top-level and handle printing *)
  let eval_top ((global_vars : v_table), (global_matrices : m_table)) (exp : expr) : rettype * global_config =
  match exp with
    Call(f, para_list) ->
        let (actuals: float list), (_, _) = List.fold_left
            (
            fun (actuals, (global_vars, global_matrices)) actual ->
              let v, (global_vars, global_matrices) = eval (global_vars, global_matrices) actual in
               match v with
                 Float(f) -> f :: actuals, (global_vars, global_matrices)
               | Matrix(m) -> raise (Failure ("Formular can not take in matrix"))
            )
            ([], (global_vars, global_matrices)) (List.rev para_list)
        in
        let (r, global_config) =  (eval (global_vars, global_matrices) exp) in
        let fcall_str = (f ^ "(" ^ (String.concat "," (List.map float2str actuals)) ^ ") = " ^  float2str (match r with Float(r) -> r | _ -> raise (Failure ("You shouldn't see this error")))) in
        (fcall_strs := fcall_str :: !fcall_strs);
        (r,global_config)
    | Binop(e1, op, e2) ->
        (match op with
        Equal | Neq | Less | Leq | Greater | Geq  ->
                let (r, global_config) = (eval (global_vars, global_matrices) exp) in
                let lvali_str = (fst(expr_to_string exp) ^ " is " ^ (match r with Float(f) -> if f > 0. then "true" else "false" | _ -> raise (Failure ("You shouldn't see this error")))) in
                (lvali_strs := lvali_str :: !lvali_strs);
                (r, global_config)
        | _ -> eval (global_vars, global_matrices) exp
        )
    | _ -> eval (global_vars, global_matrices) exp
  in


  let process_stmt (global_config:global_config) (statement : statement) =
  match statement with
  	Seq(expr_list) -> (List.fold_left
  		(fun a b -> snd(eval_top a b)) global_config expr_list)
  in

  (* Final Report Printing *)
  let _=  print_endline (title ^ " " ^ author) in
  let _ = print_endline "----------------------------------------\n         Formular Definitions\n----------------------------------------" in
  let _ = List.iter (fun f -> print_endline(formular_to_string_interpreter f)) formulars in
  let (global_vars, global_matrices) = List.fold_left process_stmt (global_vars, global_matrices) statements in
  let _ = print_endline "----------------------------------------\n         Formular Evaluation\n----------------------------------------" in
  let _ = List.iter print_endline (List.rev !fcall_strs) in
  let _ = print_endline "----------------------------------------\n         Logical Validation\n----------------------------------------" in
  let _ = List.iter print_endline (List.rev !lvali_strs) in
  let _ = print_endline "----------------------------------------\n         Variable Definitions\n----------------------------------------" in
  StringMap.iter (fun key value -> print_endline(key ^ " = " ^ float2str(value))) global_vars;
  print_endline "----------------------------------------\n         Matrix Definitions\n----------------------------------------";
  StringMap.iter (fun key value -> print_endline(key ^ " = " ^ Ast.matrix_decl_s(value))) global_matrices

